;;; -*- Mode: Lisp -*-
;;; Local-Time - Constants, variables, conditions, and structure classes
;;; Copyright (c) 2002-2009 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :ystok.local-time)

(defconstant +secs-per-day+ (* 3600 24) "number of seconds in a day")
(defconstant +msecs-per-day+ (* +secs-per-day+ 1000) "number of milliseconds in a day")

;;; Currently used timezone designators:
;;;  T	     Current system timezone.
;;;  integer Session time zone displacement in seconds a la SQL-99; 
;;;	     makes no difference in daylight saving.
;;;  NIL     "Don't bother" mode.
;;;  Zone object or name - not implemented.
(defvar *session-timezone* t
  "Session timezone designator as well as a default for all time operations")

(defvar *default-encode-type* 'local-time
  "Representation class used for encoding when timezone is not specified explicitly
 and is not null")

;;; Specials to convert text to local-time when only time components are entered
(defvar *default-year* nil)
(defvar *default-month* nil)
(defvar *default-day* nil)
;(defvar *default-date* nil)

(defvar *internal-separator* #\T)	; or #\Space

(defvar *month-keywords* #(:january :february :march :april :may :june
                           :july :august :september :october :november :december))

(defun month-name (month)
  (svref *month-names* (1- month)))

(defun ordinal-month (month-keyword)
  "Return the zero-based month number for the given MONTH keyword."
  (position month-keyword *month-keywords*))

(defparameter *weekday-keywords*
  #(:monday :tuesday :wednesday :thursday :friday :saturday :sunday))

;(defmacro dow-name (dow)
;  `(svref *weekday-names* ,dow))

(defun ordinal-day (day-keyword)
  "Return the zero-based day number for the given DAY keyword."
  (position day-keyword *weekday-keywords*))

;; Timezones
;(defvar *timezones* (make-hash-table :test #'equal) "Timezones hashed by ID")
;(defparameter *UTC* nil) ;"UTC"

;;; Conditions
(eval-when (:compile-toplevel #+debug :load-toplevel)
 (yl:defnative "Bad~@[ ~A~] fragment of local-time: ~S"
     #+Russian "~@[  ~A~]       : ~S")
 (yl:defnative "Bad~@[ ~A~] fragment of duration: ~S"
     #+Russian "~@[  ~A~]     : ~S")
)

(define-condition local-time-error (error)
 ;; year, month or whatever
 ((component-name :initarg :component-name :reader component-name :initform nil)
  (bad-token :initarg :bad-token :reader bad-token))
 (:report (lambda (c stream)
	     (format stream #L"Bad~@[ ~A~] fragment of local-time: ~S"
		     (component-name c) (bad-token c)))))

(define-condition local-time-parse-error (local-time-error parse-error) ())
(define-condition iso-8601-parse-error (local-time-error parse-error) ())

(define-condition duration-parse-error (local-time-error parse-error) ()
 (:report (lambda (c stream)
	     (format stream #L"Bad~@[ ~A~] fragment of duration: ~S"
		     (component-name c) (bad-token c)))) )

;;; Classes

;; Slot day hold a value interpreted depending on *session-timezone*
;; i.e. with or without taking DST into account.
;; SQL-99: DATE
(defstruct (date (:constructor make-date)
                 (:conc-name local-time-))
  (day 0))

;; Slots day and sec hold values interpreted depending on *session-timezone*
;; i.e. with or without taking DST into account
;; SQL-99: TIME WITHOUT TIME ZONE, TIMESTAMP WITHOUT TIME ZONE, 
(defstruct (local-time (:include date)
                       (:constructor %make-local-time))
  ;(day 0)
  (sec 0 :type fixnum)
  (msec 0 :type fixnum))

;; Slots store UTC values independent on *session-timezone* and DST
;; SQL-99: TIME WITH TIME ZONE, TIMESTAMP WITH TIME ZONE
(defstruct (utc-time (:include local-time)
                     (:conc-name local-time-)
                     (:constructor %make-utc-time))
  (zone 0))			; number of seconds representing zone displacement

;;; Duration abstract
(defstruct (duration (:constructor nil)
                     (:copier nil)))

;;; SQL-99 YEAR-MONTH INTERVAL: year unconstrained, month 0-11
(defstruct (ym-duration (:include duration)
                         (:conc-name duration-)
                         (:constructor %make-ym-duration))
   (month 0 :type fixnum))	; years * 12 + months

;;; SQL-99 DAY-TIME INTERVAL: day unconstrained; hours + minutes + seconds 0-59.999
(defstruct (dt-duration (:include duration)
                         (:conc-name duration-)
                         (:constructor %make-dt-duration))
   (day 0 :type fixnum)
   (sec 0 :type fixnum)		; hours * 3600 + minutes * 60 + seconds
   (msec 0 :type fixnum))	; 0-999


;;; Widely used fundamental utils

(declaim (ftype (function (fixnum) t) leap-year-p)
         (inline leap-year-p))
(defun leap-year-p (year)
  "Value: T if YEAR is a leap yeap in the Gregorian calendar"
  (declare (type fixnum year)
           (optimize (speed 3) #+lispworks (hcl:fixnum-safety 0)))
  (and (= 0 (the fixnum (mod year 4)))
       (or (/= 0 (the fixnum (mod year 100)))
           (= 0 (the fixnum (mod year 400))))))

(declaim (ftype (function (fixnum fixnum &optional t) fixnum) days-in-month))
(defun days-in-month (year month &optional (errorp t))
  "the number of days in MONTH of YEAR, observing Gregorian leap year rules"
  (declare (type fixnum month year)
           (optimize (speed 3) #+lispworks (hcl:fixnum-safety 0)))
  (cond ((eql month 2)                     ; feb
         (if (leap-year-p year) 29 28))
        ((<= 1 month 12)
         (svref #(31 28 31 30 31 30 31 31 30 31 30 31) (1- month)))
        (errorp
         (error (if (eq errorp t) 'local-time-error errorp)
                :component-name :month :bad-token month))
        (0)))

(declaim (ftype (function (fixnum fixnum fixnum &optional t) t) valid-date-p))
(defun valid-date-p (year month day &optional errorp)
  "t if DATE (year month day) exists in the Gregorian calendar"
  ;; Args: errorp := nil | condition type | t (means local-time-error)
  (declare (type fixnum year month day)
           (optimize (speed 3) #+lispworks (hcl:fixnum-safety 0)))
  (cond ((<= 1 day (days-in-month year month errorp)))
        (errorp 
         (error (if (eq errorp t) 'local-time-error errorp)
                :component-name :day :bad-token day))))

(defun valid-ywd-date-p (year week day &optional errorp)
  "t if DATE (year week day) exists in the Gregorian calendar"
  ;; Args: errorp := nil | condition type | t (means local-time-error)
  (declare (ignore year)
           (type fixnum week month day)
           (optimize (speed 3) #+lispworks (hcl:fixnum-safety 0)))
  (cond ((or (< week 1) (< 53 week))
         (when errorp 
           (error #1=(if (eq errorp t) 'local-time-error errorp)
                  :component-name :week :bad-token week)))
        ((<= 1 day 7))
        (errorp 
         (error #1# :component-name :day :bad-token day))))

(defmacro month3-doy (month3)
 ;;; Compute the index of the first day of month3: 0 <= month3 < 12, 0=Mar,...,11=Feb
  `(the fixnum (svref '#(0 31 61 92 122 153 184 214 245 275 306 337) ,month3)))

(declaim (ftype (function (fixnum fixnum fixnum &optional t) t) day-of-year))
(defun day-of-year (year month day &optional (check t))
 "The day number within the year of a Gregorian date, e.g. 1987 1 1 returns 1"
  (declare (type fixnum year month day)
           (optimize (speed 3) #+lispworks (hcl:fixnum-safety 0)))
  (when (or (not check) (valid-date-p year month day t))
    (if (< month 3)	
        (the fixnum (+ day (the fixnum (if (= month 1) 0 31))))
        (the fixnum (+ day (month3-doy (- month 3))
           (the fixnum (if (leap-year-p year) #.(+ 31 29) #.(+ 31 28))))))))
