;;; -*- Mode: Lisp -*-
;;; Local-Time - Encoding, decoding, and conversion
;;; Copyright (c) 2002-2009 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :ystok.local-time)

;;; Low-level local to Gregorian date conversion table and various internals

(defvar *lt-ymd-conversion-table* nil)
(defconstant +days-per-century+ 36524)
(defconstant +days-per-lt-cycle+ 146097)

(declaim (ftype (function (fixnum fixnum fixnum &optional t) fixnum)
                %ymd-lt-day %ywd-lt-day))
(defun %ymd-lt-day (year month day &optional check)
 "Compute local-time epoch day number corresponding to Gregorian date"
  (declare (type fixnum year month day)
           (optimize (speed 3) #+lispworks (hcl:fixnum-safety 0)))
  (if (or (not check) (valid-date-p year month day t))
      (let ((dy (if (< month 3) (- year 2001) (- year 2000)))
            (month3 (if (< month 3) (+ month 9)		; month3 10..11
                                    (- month 3))))	; month3 0..9
        (declare (fixnum dy month3))
        (the fixnum (+ (the fixnum (1- day))
                       (month3-doy month3)
                       (the fixnum (* 365 dy))			 ; days in prior years
                       (the fixnum (floor dy 4))		 ; Julian leap years
                       (the fixnum (- (the fixnum (floor dy 100)))); century years
                       (the fixnum (floor dy 400)))))		 ; Gregorian leap years
      most-negative-fixnum))

(defun %build-ymd-conversion-table ()
 ;;; Conversion table for one 100-year cicle.
  (declare (optimize (speed 1) (space 3) #-debug (safety 0)
                     #+(and lispworks (not debug)) (hcl:fixnum-safety 0)))
  (flet ((ymd-to-lt-entry (year month day)
           (let ((elt 0))
             (declare (type (unsigned-byte 16) elt))
             (setf (ldb (byte 7 0) elt) year
                   (ldb (byte 4 7) elt) month
                   (ldb (byte 5 11) elt) day)
             elt)))
    (loop with table = (make-array +days-per-century+ :element-type '(unsigned-byte 16))
          and year fixnum = 0
          and month fixnum = 3
          and day fixnum = 1
          with days-in-month fixnum = (days-in-month year month nil)
          for index upfrom 0 below +days-per-century+
          do (setf (aref table index) (ymd-to-lt-entry year month day))
          when (> (incf day) days-in-month)
            do (setf day 1)
            and when (> (incf month) 12)
                  do (setf month 1)
                     (incf year)
                end
            and do (setf days-in-month (days-in-month year month nil))
            finally (setq *lt-ymd-conversion-table* table)))
  t)
;; (hcl:find-object-size local-time::*lt-ymd-conversion-table*) => 73216

(eval-when (:load-toplevel :execute) (%build-ymd-conversion-table))

(defun %ywd-lt-day (year week day &optional check)
 ;;; Args: ISO-8061 year, calendar week number 1..53, and day number
  (declare (type fixnum year week day)
           (optimize (speed 3) (safety 0) #+lispworks (hcl:fixnum-safety 0)))
  (if (or (not check) (valid-ywd-date-p year week day t))
      (let ((jan-day (%ymd-lt-day year 1 4 nil)))	; January 4th
        (declare (fixnum jan-day))
        (the fixnum	; local-time-day value of the Monday of the calendar week number 1
             (+ (the fixnum (- jan-day (the fixnum (mod (the fixnum (+ jan-day 2)) 7))))
                (the fixnum (* (the fixnum (1- week)) 7))
                (the fixnum (1- day)))))
      most-negative-fixnum))

;;; Normalization accept 3 values representing the components of a local-time
;;; instance, and normalize them so that the seconds and milliseconds
;;; fall within their specified ranges:
;;;	msec:	0 ... 999
;;;	sec:	0 ... 86399
;;;	day: -infinity ... +infinity
;;; Return the three new values

#|(defun %normalize-lt-slots (day sec msec)
  (multiple-value-bind (msec-seconds msecs)
      (floor msec 1000)
    (multiple-value-bind (sec-days seconds)
        (floor (+ msec-seconds sec) +secs-per-day+)
      (values (+ day sec-days) seconds msecs))))|#

(defmacro with-normalized-lt-day-sec (((var-day day) (var-sec sec)
                                       &optional ((&optional var-msec msec)))
                                      &body body)
 ;;; Args: (var-msec msec)  Is not needed for operations like zone adjusting.
  (with-gensyms (sec-days)
    (if msec
        (with-gensyms (msec-seconds)
          `(multiple-value-bind (,msec-seconds ,var-msec) (floor ,msec 1000)
             (declare (fixnum ,msec-seconds ,var-msec))
             (multiple-value-bind (,sec-days ,var-sec)
                 (floor (+ ,sec ,msec-seconds) +secs-per-day+)
               (declare (fixnum ,sec-days ,var-sec))
               (let ((,var-day (the fixnum (+ ,day ,sec-days))))
                 (declare (fixnum ,var-day))
                 ,@body))) )
        `(multiple-value-bind (,sec-days ,var-sec) (floor ,sec +secs-per-day+)
           (declare (fixnum ,sec-days ,var-sec))
           (let ((,var-day (the fixnum (+ ,day ,sec-days))))
             (declare (fixnum ,var-day))
             ,@body)) )))

;;; Decoders for normalized day and sec

(declaim (ftype (function (fixnum) (values fixnum fixnum fixnum))
                %lt-day-ymd %lt-sec-hms)
         (ftype (function (fixnum) fixnum) %lt-day-dow %lt-day-week))
(defun %lt-day-ymd (local-time-day)
 "Values: Greagorian year, month and day"
  ;; NB: This function has replaced %gregorian-date, 
  ;;     and we have opencoded the former lt-entry-to-ymd.
  (declare (fixnum local-time-day)
           #-debug (optimize (safety 0) (speed 3) #+lispworks (hcl:fixnum-safety 0)))
  (multiple-value-bind (cycle cycle-day) (floor local-time-day +days-per-lt-cycle+)
    (declare (fixnum cycle cycle-day))
    (if (= cycle-day (1- +days-per-lt-cycle+))		    ; last cycle day falls 
        (values (the fixnum (+ 2400 			    ; outside 4 centuries and
                               (the fixnum (* cycle 400)))) ; implies century is 4
                2				
                29)    
        (multiple-value-bind (century century-day) (floor cycle-day +days-per-century+)
          (declare (fixnum century century-day))
          (let ((elt (aref *lt-ymd-conversion-table* century-day)))
            (declare (type (unsigned-byte 16) elt))
            (values (the fixnum (+ 2000
                                   (the fixnum (* cycle 400))
                                   (the fixnum (* century 100))
                                   (ldb (byte 7 0) elt)))
                    (ldb (byte 4 7) elt)
                    (ldb (byte 5 11) elt)))))))

#|(defun lt-entry-to-ymd (elt)
  (declare (type (unsigned-byte 16) elt))
  (values (ldb (byte 7 0) elt) (ldb (byte 4 7) elt) (ldb (byte 5 11) elt)))|#

;;; Next quick accessors should be used cautiously because of:
;;;  utc-time => week day or time according UTC
;;;  local-time => week day or time disregarding *session-timezone*
(defun %lt-day-dow (local-time-day)
 ;;; Value: day of week number: 0=Mon,...,6=Sun etc.
  ;; NB: Coincides with Common Lisp, where 0=Mon,...,6=Sun.
  ;;     Differs from Unix or Windows weekday encoding, 0=Sun,1=Mon,...,6=Sat
  (declare (fixnum local-time-day) #+lispworks (optimize (hcl:fixnum-safety 0)))
  (mod (the fixnum (+ local-time-day 2)) 7))

(defun %lt-day-week (local-time-day)
 ;;; Value: Calendar week number within the year in 0..53.
  ;; The first calendar week is the week in which January 1 occurs.
  (declare (optimize (speed 3) (safety 0) #+lispworks (hcl:fixnum-safety 0))
           (fixnum local-time-day))
  (let ((jan-day (%ymd-lt-day (%lt-day-ymd local-time-day) 1 1 nil)))	; January 1st
          ;(first-monday-day (- jan-1-day (%lt-day-dow jan-1-day)))
    (declare (fixnum jan-day))
    (the fixnum				; get local-time-day value of the Monday
         (1+ (the fixnum		; of the first week of the same year.
                  (floor (the fixnum (- local-time-day
                                        (the fixnum (- jan-day
                           (the fixnum (mod (the fixnum (+ jan-day 2)) 7)))))) 7))))))

(defun %lt-day-week-iso (local-time-day)
 ;;; Values: 1) calendar week number 1..53 within the year that may not coincide.
  ;;	     2) actual year	
  ;; The first calendar week is the week which includes January 4,
  ;; or, to put another way, which has at least four days in the year.
  (declare (optimize (speed 3) (safety 0) #+lispworks (hcl:fixnum-safety 0))
           (fixnum local-time-day))
  (let* ((year (%lt-day-ymd local-time-day))
         (jan-day (%ymd-lt-day year 1 4 nil))	; January 4th
         (week (the fixnum			; get local-time-day value of the Monday
          (1+ (the fixnum			; of the first week of the same year.
                  (floor (the fixnum (- local-time-day
                                        (the fixnum (- jan-day
                           (the fixnum (mod (the fixnum (+ jan-day 2)) 7)))))) 7))))))
    (declare (fixnum year jan-day week))
    (cond ((= week 0)           	; replace by 52 or 53 of the previous year
           (%lt-day-week-iso (- jan-day 4)))	;=(%ymd-lt-day (1- year) 12 31)))
          ((= week 53)			; check if it is week 1 of the following year
           (if (< (%lt-day-dow (%ymd-lt-day year 12 31)) 3)
               (values 1 (1+ year))		 ; Mon, Tue, Wed => 1
               (values 53 year)))	
          (t (values week year)))))

(defun %lt-sec-hms (local-time-sec)
 ;;; Values: hour, minute, second.
  ;; NB: If the sec value is not normalized, can return hours outside [0...23]
  (declare (fixnum local-time-sec) #+lispworks (optimize (hcl:fixnum-safety 0)))
  (multiple-value-bind (minutes second) (floor local-time-sec 60)
    (declare (fixnum minutes second))
    (multiple-value-bind (hour minutes) (floor minutes 60)
      (declare (fixnum hour minutes))
      (values hour minutes second))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Encoding

(defun make-local-time (&key (day 0) (sec 0) (msec 0) zone)
 "Make a local-time instance"
  ;; NB: Do we really need to normalize msec?
  ;(multiple-value-bind (%day %sec %msec) (%normalize-lt-slots day sec msec)
  ;(multiple-value-bind (msec-seconds msec) (floor msec 1000)
  ;  (declare (fixnum msec))
    ;(multiple-value-bind (sec-days seconds) (floor (+ msec-seconds sec) +secs-per-day+)
  (with-normalized-lt-day-sec ((day day) (sec sec) (msec msec)) ; (+ msec-seconds sec)))
    (if zone
        (%make-utc-time   :day day  :sec sec  :msec msec  :zone zone)
        (%make-local-time :day day  :sec sec  :msec msec))))

(declaim (ftype (function (fixnum fixnum fixnum &optional t) date) encode-date)
         (inline encode-date))
(defun encode-date (year month day &optional (check t))
  (declare (fixnum day month year))
  (make-date :day (if (minusp month)
                      (%ywd-lt-day year (- month) day check)
                      (%ymd-lt-day year month day check))))

(defun encode-local-time (ms ss mm hh day month year
                          &optional (timezone *session-timezone*)
                                    (encode-type *default-encode-type*)
                                    (check t))
 "Return a new local-time instance corresponding to the specified time elements"
  ;; Args: timezone 
  (when (and check (not (valid-date-p year month day t)))	; should not be (month<0)
    (return-from encode-local-time nil))
  (cond ((neq encode-type 'utc-time)
         #1=(make-local-time :day #2=(if (minusp month)
                                         (%ywd-lt-day year (- month) day) ; week number
                                         (%ymd-lt-day year month day))	  ; check:=nil
                             :sec (+ ss (* mm 60) (* hh 3600))
                             :msec ms
                             :zone nil))
            ;(if (eq timezone t) (%unix-to-local-time (+ unix-time zone) ms nil))
        ((eq timezone t)        
             ;; Let run-time library determine displacement;
             ;; verify correctness as mktime does not check components.
             ;(let* ((unix-time (unix-mktime (- year 1900) (1- month) day hh mm))
             ;       (zone (unix-timezone unix-time)))
             ;  (%unix-to-local-time unix-time ms zone)
         ;; Do we need to normalize hh before calling %ymd-timezone??
         (when (minusp month)
           (multiple-value-setq (year month day)
               (%lt-day-ymd (%ywd-lt-day year (- month) day))))	; to compute timezone
         (setq timezone (%ymd-timezone year month day hh))
         (make-local-time :day (%ymd-lt-day year month day)	; check:=nil
                          :sec (+ ss (* mm 60) (* hh 3600) (- timezone))
                          :msec ms
                          :zone timezone))
        ((numberp timezone)
         (make-local-time :day #2#
                          :sec (+ ss (* mm 60) (* hh 3600) (- timezone))
                          :msec ms
                          :zone timezone))
        (t #1#)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Decoding
;;; decode-as-local and decode-as-utc must not be applied to a date instance!

(defmacro with-adjusted-hms ((hour-days hours minutes seconds) local-time-sec
                             &body body)
 ;;; Bind the variables according to local-time-sec. Hour-days - adjustment
  `(multiple-value-bind (,minutes ,seconds) (floor ,local-time-sec 60)
    (declare (fixnum ,minutes ,seconds) (ignorable ,seconds))
    (multiple-value-bind (,hours ,minutes) (floor ,minutes 60)
      (declare (fixnum ,hours ,minutes) (ignorable ,minutes))
      (multiple-value-bind (,hour-days ,hours) (floor ,hours 24)
        (declare (fixnum ,hour-days ,hours) (ignorable ,hours))
        ,@body))))

(declaim (ftype (function (date) (values fixnum fixnum fixnum)) decode-date)
         (inline decode-date))
(defun decode-date (date)
 "Return multiple values: the normal year, month, and day of the date or local-time"
  ;; May also be used to truncate time components from the argument, an instance of
  ;;  local-time => local midnight
  ;;  utc-time   => UTC midnight
  (%lt-day-ymd (local-time-day date)))

(defun decode-as-local (local-time &optional timezone-p)
 "Returns the decoded components of as locally related multiple values:
 msec, seconds, minutes, hours, day, month, year, week-day-number, 
 timezone, daylight-saving-delta."
 ;;; Arguments:
  ;; TIMEZONE-P	If null, timezone and daylight-saving-delta are not returned
  ;;		if this requires a considerable computation effort.
  ;; Values:
  ;;  WEEK-DAY-NUMBER is 1-based: 1=monday, etc.
  ;;  TIMEZONE  is a total displacement, either standard or dst as appropriate, in secs;
  ;;  DAYLIGHT SAVING is displacement increment or decrement in seconds correspondingly.
  ;; NB1: The order timezone and daylight-saving-delta values returned is different
  ;;      from (decode-universal-time).
  ;; NB2: The daylight-saving-delta or both of them may be returned as NIL.
  ;;      If local-time is utc-time-p, return NIL as a daylight-saving-delta 
  ;;      and total displacement as timezone.
  ;;      The daylight-saving-delta actually is not used by any caller.
  ;; NB3: The customary timezone abbreviation is not implemented.
  (let (zone dst)
    (flet ((decode (local-time displacement)
             (declare (fixnum displacement))
             (with-adjusted-hms (hour-days hours minutes seconds)
                 (+ (local-time-sec local-time) displacement)
               (let ((local-time-day (+ (local-time-day local-time) hour-days)))
                 (multiple-value-bind (year month day) 
                     (%lt-day-ymd local-time-day)
                   (values (local-time-msec local-time)
                           seconds minutes hours day month year
                           (1+ (%lt-day-dow local-time-day))
                           zone dst))))))
      (cond ((utc-time-p local-time)
             (decode local-time (setq zone (local-time-zone local-time))))
            ((and (eq *session-timezone* t) timezone-p)
             (multiple-value-setq (zone dst) (%lt-timezone (local-time-day local-time)
                                                           (local-time-sec local-time)))
             ;; Interpret relative the session timezone
             ;; Try to take the dst value into play (SQL-99 does not!)
             ;(multiple-value-bind (zone dst-p) (unix-timezone (unix-time local-time))
             (decode local-time 0))
            (t
             (when (numberp *session-timezone*) (setq zone *session-timezone*))
             (decode local-time 0))))))

(defun decode-as-utc (local-time)
 "Returns the decoded components as UTC-related multiple values:
 mecs, seconds, minutes, hours, day, month, year"
  (flet ((decode (local-time displacement)
           (declare (fixnum displacement))
           (with-adjusted-hms (hour-days hours minutes seconds)
               (- (local-time-sec local-time) displacement)
             (multiple-value-bind (year month day) 
                 (%lt-day-ymd (+ (local-time-day local-time) hour-days))
               (values (local-time-msec local-time)
                       seconds minutes hours day month year)))))
    (cond ((utc-time-p local-time)		; No adjustment needed
           (decode local-time 0))
          ((eq *session-timezone* t)
           (decode local-time (%lt-timezone (local-time-day local-time)
                                            (local-time-sec local-time))))
           ;(decode local-time (unix-timezone (unix-time local-time))))
           ;(let ((unix-time (unix-time local-time)))
           ;  (multiple-value-bind (seconds minutes hours day month year wday yday dst)
           ;      (unix-gmtime unix-time)
           ;    (declare (ignore yday))
           ;    (values (local-time-msec local-time)
           ;            seconds minutes hours day (1+ month) (+ 1900 year)))))
          (t (decode local-time
                     (if (numberp *session-timezone*) *session-timezone* 0))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Adjusting
;;; lt -> utc: utc.sec := lt.sec  - lt.zone
;;; utc -> lt: lt.sec  := utc.sec + utc.zone

(defun local-time-adjust (source timezone &optional destination)
 ;;; Returns two values, the values of new day and sec slots, 
  ;; for reconstructing a utc-time instance by means of make-local-time.
  ;; If destination is a local-time instance, fills the slots
  ;; with the new values and returns the destination.
  ;; NB: Generally used for printing local-time instances as of the timezone given
  ;;	 as simply rebinding *session-timezone* seems not of any help.
  (let ((day (local-time-day source))
        (sec (local-time-sec source)))
    (declare (fixnum day sec))
    (if destination
        ;; canonicalize destination zone: use standard or current (win32-timezone)?
        (let ((timezone (if (eq timezone t) *standard-timezone* timezone)))
          (declare (fixnum timezone))
          (if (utc-time-p source)
              (if (utc-time-p destination)
                  ;; utc -> utc: simply change timezone
                  (setf (local-time-day destination) day
                        (local-time-sec destination) sec
                        (local-time-zone destination) timezone)
                  ;; utc -> lt: convert to timezone 
                  ;; Most dubious operation: 
                  ;; - have some sense only for timezone = t,
                  ;; - does not take dst-p into account!
                  (with-normalized-lt-day-sec ((day day)
                                               (sec (+ sec timezone)))
                    (setf (local-time-day destination) day
                          (local-time-sec destination) sec)) )
              (let ((source-zone (timezone source)))
                (if (utc-time-p destination)
                    ;; lt -> utc: interpret lt as of the *session-timezone*
                    (with-normalized-lt-day-sec ((day day)
                                                 (sec (- sec source-zone)))
                      (setf (local-time-day destination) day
                            (local-time-sec destination) sec
                            (local-time-zone destination) timezone))
                    ;; lt -> lt: adjust according zone difference
                    (with-normalized-lt-day-sec ((day day)
                                                 (sec (+ sec (- timezone source-zone))))
                      (setf (local-time-day destination) day
                            (local-time-sec destination) sec)) )))
          destination)
      ;; Return values for constructing a utc-time instance by means of make-local-time
      (with-normalized-lt-day-sec ((day day)
                                   (sec (if (utc-time-p source)
                                            sec
                                            (- sec (timezone source)))))
        (values day sec)))))

(defun local-timezone (adjusted-local-time &optional (timezone *session-timezone*))
 ;;; Return the local timezone adjustment applicable at the already adjusted-local-time.
  ;; Used to reverse the effect of timezone and local-time-adjust.
  ;; canonicalize destination zone: use standard or current (win32-timezone)?
  (let ((timezone (if (eq timezone t) *standard-timezone* timezone)))
    (if (utc-time-p adjusted-local-time)
        (- (local-time-zone adjusted-local-time) timezone)
        0)))	; have any sense?

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Conversion between time systems

(defconstant +universal-to-local+ 3160857600)
(defconstant +unix-to-local+ 951868800) ;(- +universal-to-local+ +unix-to-universal+)
;(defconstant +unix-to-universal+ 2208988800)

(defun %universal-to-local-time (universal-time &optional (msec 0) (zone 0))
  "Convert universal timeval to local-time instance."
  (make-local-time :sec (- universal-time +universal-to-local+) :msec msec :zone zone))

(defun universal-time (local-time)
 "Return the universal-time corresponding to the local-time"
  ;; Should be applicable to dates?
  (values (+ (* (local-time-day local-time) +secs-per-day+)
             (cond ((utc-time-p local-time)
                    (local-time-sec local-time))
                   ((local-time-p local-time)
                    (- (local-time-sec local-time) (timezone local-time)))
                   ((- (timezone local-time))))	; assume date is always local
             +universal-to-local+)
          (local-time-msec local-time)))

(defun %unix-to-local-time (unix-time &optional (msec 0) (zone 0))
  "Convert unix timeval to local-time instance."
  ;(%universal-to-local-time (+ unix +unix-to-universal+) msec)
  (make-local-time :sec (- unix-time +unix-to-local+) :msec msec :zone zone))

(defun unix-time (local-time)
 ;; Usually deals with local-time instances rather than utc-time.
 ;; To Do: Find a sound solution to get timezone displacement if not utc-time.
 "Return the unix time corresponding to the local-time"
 (values (+ (* (local-time-day local-time) +secs-per-day+)
            (cond ((utc-time-p local-time)
                   (local-time-sec local-time))
                  ((local-time-p local-time)
                   (- (local-time-sec local-time) (timezone local-time)))
                  ((- (timezone local-time))))	; assume date is always local
            +unix-to-local+)
         (local-time-msec local-time)))

(defun internal-time (local-time)
  "return the internal system time corresponding to the local-time"
  (declare (ignore local-time))
  (error "not implemented"))

(defun local-time (&key universal internal unix (msec 0) (zone 0))
  "Produce a local-time instance from the provided numeric time representation"
 ;; Args: zone - This times usually represents UCT
 ;; NB: The source value is interpreted as of the zone given
  (cond (unix (%unix-to-local-time unix msec zone))
        ;(java (%java-to-local-time java msec))
        (internal (error ":internal argument unsupported"))
        (universal (%universal-to-local-time universal msec zone))
        (t
         (error "Specify one of :universal, :java, :internal or :unix keywords"))))

(pushnew :local-time *features*)


#||;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; EVALUATION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#+unused
(defun decode-gmtime (local-time)
  (let ((unix-time (unix-time local-time)))
    (multiple-value-bind (seconds minutes hours day month year wday yday dst)
        (unix-gmtime unix-time)
      (declare (ignore yday))
      (values (local-time-msec local-time)
              seconds minutes hours day (1+ month) (+ 1900 year) wday
              (= 1 dst) (unix-tz-displacement unix-time) "Not Implemented"))))

(defun local-time-hours (local-time)
  "return the local-time as a number of hours, rounded down"
  (let* ((day-hours (* 24 (local-time-day local-time))))
    (+ day-hours
       (floor (/ (local-time-sec local-time) 3600)))))

(defconstant +java-to-unix+ 1000)
(defun %java-to-local-time (java &optional (msec2 0))
  "Convert java timeval to local-time instance. Java is unix * 1000"
  (multiple-value-bind (sec msec) (floor java +java-to-unix+)
    (%unix-to-local-time sec (+ msec msec2))))
(defun %local-time-to-java (local-time)
  (multiple-value-bind (unix msec) (%local-time-to-unix local-time)
    (+ (* unix +java-to-unix+) msec)))
||#

