;;; -*- Mode: Lisp -*-
;;; Local-Time - Duration core and operations
;;; Copyright (c) 2002-2009 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; For a sign-normalized duration, all slots are either >=0 or <=0.
;;; Sign-normalization is needed for printing, comparing by =, <, <=.
;;; This is not needed for arithmetic operations.

(in-package :ystok.local-time)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Make and encode

;;; year-month duration instance
(defun encode-ym-duration (year month &optional (check t))
  (cond ((null check))
        ((not (integerp year)) (error 'duration-parse-error
                                      :component-name :year  :bad-token year))
        ((not (integerp month)) (error 'duration-parse-error
                                       :component-name :month  :bad-token month)))
  (%make-ym-duration :month (the fixnum (+ month (* year 12)))))

;;; day-second duration instance

(defmacro with-normalized-ds-day-sec-msec (((var-day day) (var-sec sec) (var-msec msec))
                                           &body body)
  (with-gensyms (msec-seconds sec-days)
    `(multiple-value-bind (,msec-seconds var-msecs)
         (truncate ,msec 1000)
       (declare (fixnum ,msec-seconds ,var-msec))
       (multiple-value-bind (,sec-days ,var-sec)
           (truncate (the fixnum (+ ,msec-seconds ,sec)) +secs-per-day+)
         (declare (fixnum ,sec-days ,var-sec))
         (let ((,var-day (the fixnum (+ ,day ,sec-days))))
           (declare (fixnum ,var-day))
           ,@body)))))

(defmacro sign-normalize-day-sec-msec (day sec msec)
 ;;; Guarantee that all slots are of the same signum
  ;; Args: day, sec, msec - variables or slot references with values already normalized
  `(cond ((> ,day 0)			; results in non-negative duration
          (cond ((< ,sec 0)
                 (when (< ,msec 0)
                   (incf ,msec 1000)
                   (decf ,sec))
                 (incf ,sec +secs-per-day+)
                 (decf ,day))
                ((< ,msec 0)
                 (incf ,msec 1000)
                 (when (< (decf ,sec) 0)
                   (incf ,sec +secs-per-day+)
                   (decf ,day))) ))
         ((< ,day 0)			; results in non-positive duration
          (cond ((> ,sec 0)
                 (when (> ,msec 0)
                   (decf ,msec 1000)
                   (incf ,sec))
                 (decf ,sec +secs-per-day+)
                 (incf ,day))
                ((> ,msec 0)
                 (decf ,msec 1000)
                 (when (> (incf ,sec) 0)
                   (decf ,sec +secs-per-day+)
                   (incf ,day))) ))
         ((> ,sec 0)			; day=0, results in non-negative duration
          (when (< ,msec 0)
            (incf ,msec 1000)
            (decf ,sec)))
         ((< ,sec 0)			; day=0, results in non-positive duration
          (when (> ,msec 0)
            (decf ,msec 1000)
            (incf ,sec))) ))

(defun make-dt-duration (&key (day 0) (sec 0) (msec 0))
 ;;; Normalize slots
  (declare (fixnum day sec msec) #+lispworks (optimize (hcl:fixnum-safety 0)))
  (with-normalized-ds-day-sec-msec ((var-day day) (var-sec sec) (var-msec msec))
    (%make-dt-duration :day var-day  :sec var-sec  :msec var-msecs)))

(declaim (ftype (function (fixnum) (values fixnum fixnum fixnum)) %dt-sec-hms))
(defun %dt-sec-hms (sec)
  (declare (fixnum sec))
  (multiple-value-bind (sec-minutes seconds) (truncate sec 60)
    (declare (fixnum sec-minutes seconds))
    (multiple-value-bind (hours minutes) (truncate sec-minutes 60)
      (declare (fixnum hours minutes))
      (values hours minutes seconds))))

(defun %get-day-sec-msec (day &optional (hours 0) (minutes 0) (seconds 0) (msec 0))
 ;;; Values: day, sec, msec - all non-normalized
  ;; NB: Using round instead of truncate for msec would imply
  ;;     - benefits: better read-print consistency;
  ;;	 - drawbacks: (decode-duration number) required normalization afterwards.
  (let ((day (if (integerp day)
                 day
                 (multiple-value-bind (d f) (truncate day)       
                   (declare (fixnum d))
                   (setq seconds (+ seconds (* f +secs-per-day+)))
                   d))))			 ; guarantee fixnum day
    (declare (fixnum day))
    (when (/= hours 0) (incf seconds (* hours 3600)))
    (when (/= minutes 0) (incf seconds (* minutes 60)))
    (if (integerp seconds)  
        (values day seconds msec)
        (multiple-value-bind (sec f) (truncate seconds)
          (declare (fixnum sec))
          (values day sec (the fixnum (truncate (+ (* f 1000d0) msec))))))))

(declaim (ftype (function (real) (values fixnum fixnum fixnum)) %number-day-sec-msec))
(defun %number-day-sec-msec (number)
 ;;; Optimized version of %get-day-sec-msec
  ;; Arg: number  Real number of days
  ;; Values: day, sec, msec - all non-normalized
  ;; NB: Using round instead of truncate for msec would imply
  ;;     benefits: better read-print consistency;
  ;;	 drawbacks: (decode-duration number) required normalization afterwards.
  (if (integerp number)
      (values number 0 0)
      (multiple-value-bind (day f) (truncate number)       
        (declare (fixnum day))
        (multiple-value-bind (sec f) (truncate (* f +secs-per-day+))
          (declare (fixnum sec))
          (values day sec (the fixnum (truncate (* f 1000d0))))))))


(defun encode-dt-duration (day &optional (hours 0) (minutes 0) (seconds 0) (msec 0)
                                         (check t))
 ;;; Args: Any number of non-integer components are allowed 
  ;;       check  If true, verify that all components are of the same sign.
  (when check
    (let ((list (list day hours minutes seconds msec)))
      ;; ensure normalization of sign
      (when-bind (pos-plus (position-if #'plusp list))
        (when-bind (pos-minus (position-if #'minusp list))
          (error 'duration-parse-error
                 :component-name (nth (max pos-plus pos-minus)
                                      '(:day :hour :minute :second :millisecond))
                 :bad-token (nth (max pos-plus pos-minus) list))))
  ) )
  (multiple-value-bind (day sec msec)
      (%get-day-sec-msec day hours minutes seconds msec)
    (declare (fixnum day sec msec))
    (make-dt-duration :day day :sec sec :msec msec)))

(defgeneric decode-duration (duration)
 (:documentation
  "Returns multiple values: day,minutes,hours,seconds,msec or year,month")

 (:method ((arg ym-duration))
  (truncate (duration-month arg) 12))

 (:method ((arg dt-duration))
  (let ((day (duration-day arg))
        (sec (duration-sec arg))
        (msec (duration-msec arg)))
    (declare (fixnum day sec msec))
    (sign-normalize-day-sec-msec day sec msec)	; is this the right place?
    (multiple-value-bind (hours minutes seconds) (%dt-sec-hms sec)
      (values day hours minutes seconds msec))))

 (:method ((arg number))			; days
  (multiple-value-bind (day sec msec) (%number-day-sec-msec arg)
    (declare (fixnum day sec msec))
    (multiple-value-bind (hours minutes seconds) (%dt-sec-hms sec)
      (values day hours minutes seconds msec))))
)

(defun designate-duration (duration)
  "returns the bignum value of milliseconds in DURATION"
  (etypecase duration
    (dt-duration
     (+ (* (duration-day duration) +msecs-per-day+)
        (* (duration-sec duration) 1000)
        (duration-msec duration)))
    (number
     (* duration +msecs-per-day+))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Comparizon and arithmetics

(defmacro with-dt ((day sec msec) dt &body body)
  `(let ((,day (duration-day ,dt))
         (,sec (duration-sec ,dt))
         (,msec (duration-msec ,dt)))
     (declare (fixnum ,day ,sec ,msec))
     ,@body))

(defmethod local-time=2 ((x ym-duration) (y ym-duration))
  (= (duration-month x) (duration-month y)))

(defmethod local-time=2 ((x dt-duration) (y dt-duration))
  (with-dt (x-day x-sec x-msec) x
    (with-dt (y-day y-sec y-msec) y
      (sign-normalize-day-sec-msec x-day x-sec x-msec)
      (sign-normalize-day-sec-msec y-day y-sec y-msec)
      (and (= x-sec y-sec)
           (= x-day y-day)
           (= x-msec y-msec)))))

(defmethod local-time=2 ((x dt-duration) (y number))
  (with-dt (x-day x-sec x-msec) x
    (multiple-value-bind (y-day y-sec y-msec) (%number-day-sec-msec y)
      (sign-normalize-day-sec-msec x-day x-sec x-msec)
      (and (= x-sec y-sec)
           (= x-day y-day)
           (= x-msec y-msec)))))

(defmethod local-time=2 ((x number) (y dt-duration))
  (multiple-value-bind (x-day x-sec x-msec) (%number-day-sec-msec x)
    (declare (fixnum x-day x-sec x-msec))
    (with-dt (y-day y-sec y-msec) y
      (sign-normalize-day-sec-msec y-day y-sec y-msec)
      (and (= x-sec y-sec)
           (= x-day y-day)
           (= x-msec y-msec)))))

(defmethod local-time=2 ((x number) (y number))
  (= x y))


(defmethod local-time<2 ((x ym-duration) (y ym-duration))
  (< (duration-month x) (duration-month y)))

(defmethod local-time<2 ((x dt-duration) (y dt-duration))
  (with-dt (x-day x-sec x-msec) x
    (with-dt (y-day y-sec y-msec) y
      (sign-normalize-day-sec-msec x-day x-sec x-msec)
      (sign-normalize-day-sec-msec y-day y-sec y-msec)
      (or (< x-day y-day)
          (and (= x-day y-day)
               (or (< x-sec y-sec)
                   (and (= x-sec y-sec)
                        (< x-msec y-msec))))))))

(defmethod local-time<2 ((x dt-duration) (y number))
  (with-dt (x-day x-sec x-msec) x
    (multiple-value-bind (y-day y-sec y-msec) (%number-day-sec-msec y)
      (declare (fixnum y-day y-sec y-msec))
      (sign-normalize-day-sec-msec x-day x-sec x-msec)
      (or (< x-day y-day)
          (and (= x-day y-day)
               (or (< x-sec y-sec)
                   (and (= x-sec y-sec)
                        (< x-msec y-msec))))))))

(defmethod local-time<2 ((x number) (y dt-duration))
  (multiple-value-bind (x-day x-sec x-msec) (%number-day-sec-msec x)
    (declare (fixnum x-day x-sec x-msec))
    (with-dt (y-day y-sec y-msec) y
      (sign-normalize-day-sec-msec y-day y-sec y-msec)
      (or (< x-day y-day)
          (and (= x-day y-day)
               (or (< x-sec y-sec)
                   (and (= x-sec y-sec)
                        (< x-msec y-msec))))))))

(defmethod local-time<2 ((x number) (y number))
  (< x y))

;;; Binary plus
(defmethod local-time+2 ((x ym-duration) (y ym-duration))
  (%make-ym-duration :month (the fixnum (+ (duration-month x) (duration-month y)))))

(defmethod local-time+2 ((x dt-duration) (y dt-duration))
  (make-dt-duration :day (the fixnum (+ (duration-day x) (duration-day y)))
                    :sec (the fixnum (+ (duration-sec x) (duration-sec y)))
                    :msec (the fixnum (+ (duration-msec x) (duration-msec y)))))

(defmethod local-time+2 ((x dt-duration) (y number))
  (multiple-value-bind (y-day y-sec y-msec) (%number-day-sec-msec y)
    (declare (fixnum y-day y-sec y-msec))
    (make-dt-duration :day (the fixnum (+ (duration-day x) y-day))
                      :sec (the fixnum (+ (duration-sec x) y-sec))
                      :msec (the fixnum (+ (duration-msec x) y-msec)))))

(defmethod local-time+2 ((x number) (y dt-duration))
  (multiple-value-bind (x-day x-sec x-msec) (%number-day-sec-msec x)
    (declare (fixnum x-day x-sec x-msec))
    (make-dt-duration :day (the fixnum (+ x-day (duration-day y)))
                      :sec (the fixnum (+ x-sec (duration-sec y)))
                      :msec (the fixnum (+ x-msec (duration-msec y))))))

(defmethod local-time+2 ((x number) (y number))
  (+ x y))

;;; Unary minus
(defmethod duration- ((x ym-duration))
  (%make-ym-duration :month (the fixnum (- (duration-month x)))))

(defmethod duration- ((x dt-duration))
  (%make-dt-duration :day (the fixnum (- (duration-day x)))
                     :sec (the fixnum (- (duration-sec x)))
                     :msec (the fixnum (- (duration-msec x)))))

(defmethod duration- ((x number))
  (- x))

;;; Binary minus
(defmethod local-time-2 ((x ym-duration) (y ym-duration))
  (%make-ym-duration :month (the fixnum (- (duration-month x) (duration-month y)))))

(defmethod local-time-2 ((x dt-duration) (y dt-duration))
  (make-dt-duration :day (the fixnum (- (duration-day x) (duration-day y)))
                    :sec (the fixnum (- (duration-sec x) (duration-sec y)))
                    :msec (the fixnum (- (duration-msec x) (duration-msec y)))))

(defmethod local-time-2 ((x dt-duration) (y number))
  (multiple-value-bind (y-day y-sec y-msec) (%number-day-sec-msec y)
    (declare (fixnum y-day y-sec y-msec))
    (make-dt-duration :day (the fixnum (- (duration-day x) y-day))
                      :sec (the fixnum (- (duration-sec x) y-sec))
                      :msec (the fixnum (- (duration-msec x) y-msec)))))

(defmethod local-time-2 ((x number) (y dt-duration))
  (multiple-value-bind (x-day x-sec x-msec) (%number-day-sec-msec x)
    (declare (fixnum x-day x-sec x-msec))
    (make-dt-duration :day (the fixnum (- x-day (duration-day y)))
                      :sec (the fixnum (- x-sec (duration-sec y)))
                      :msec (the fixnum (- x-msec (duration-msec y))))))

(defmethod local-time-2 ((x number) (y number))
  (- x y))

;;; Binary times
(defmethod duration*2 ((x ym-duration) (y number))
  (%make-ym-duration :month (the fixnum (round (* (duration-month x) y)))))

(defmethod duration*2 ((x number) (y ym-duration))
  (%make-ym-duration :month (the fixnum (round (* x (duration-month y))))))

(defmethod duration*2 ((x dt-duration) (y number))
  (encode-dt-duration (* (duration-day x) y) 0 0
                      (* (duration-sec x) y) (* (duration-msec x) y) nil)) ; check

(defmethod duration*2 ((x number) (y dt-duration))
  (encode-dt-duration (* x (duration-day y)) 0 0
                      (* x (duration-sec y)) (* x (duration-msec y)) nil)) ; check

(defmethod duration*2 ((x number) (y number))
  (* x y))

(defun duration* (arg &rest args)
 "Multiply one duration by numbers; 
 only a single duration instance is allowed among ARG and ARGS."
  (reduce #'duration*2 args :initial-value arg))

;;; Binary quotient
(defmethod duration/2 ((x ym-duration) (y ym-duration))
  (/ (duration-month x) (duration-month y)))

(defmethod duration/2 ((x ym-duration) (y number))
  (%make-ym-duration :month (the fixnum (round (/ (duration-month x) y)))))

(defmethod duration/2 ((x dt-duration) (y dt-duration))
  ;(/ (duration-sec x) (duration-sec y)))
  (/ (designate-duration x) (designate-duration y)))

(defmethod duration/2 ((x dt-duration) (y number))
  (encode-dt-duration (/ (duration-day x) y) 0 0
                      (/ (duration-sec x) y) (/ (duration-msec x) y) nil)) ; check

(defmethod duration/2 ((x number) (y number))
  (/ x y))

(defmethod duration/2 (x y)
  (/ (designate-duration x) (designate-duration y)))

(defun duration/ (arg &rest args)
  "Divide the duration ARG by another duration or by numbers.
 Return a (rational) number or a new duration instance."
  (if args
      (reduce #'duration/2 args :initial-value arg)
      (/ arg)))


#||;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; EVALUATION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(decode-duration (encode-ym-duration -3 2))
(setq ym (encode-ym-duration -2))
(encode-ym-duration 3.2)

(setq ds (make-dt-duration :sec -600))
(setq ds (make-dt-duration :day -4 :sec 3600 :msec 100)) ;=> @P-3D22H59M59.9S
(setq ds (make-dt-duration :day -4 :sec 3600 :msec -100)) ;=> @P-3D23H0.1S 
(setq ds (make-dt-duration :sec 3600 :msec -100))

(setq ds (encode-dt-duration 10 3 2 30.5 100))

(format-duration *standard-output* ds)
||#
