;;; -*- Mode: Lisp -*-
;;; Local-Time - Calls to C run-time library and Win32 API library
;;; Copyright (c) 2002-2009 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :ystok.local-time)

#-(and lispworks win32)
(eval-when (:compile-toplevel :load-toplevel :execute)
  (error "This file can only be used on LispWorks for Windows!"))

(declaim (optimize (speed 3) #+debug (debug 3)))

(eval-when (:compile-toplevel :load-toplevel :execute)
 (defparameter *c-rtl-module* "msvcrt")
 (fli:define-c-typedef time-t :long)
 (fli:define-c-struct libc-time
    (tm-sec   :int)			; seconds
    (tm-min   :int)			; minutes
    (tm-hour  :int)			; hours
    (tm-mday  :int)			; day of month
    (tm-mon   :int)			; month
    (tm-year  :int)			; year
    (tm-wday  :int)			; day of week
    (tm-yday  :int)			; day of year
    (tm-isdst :int))			; daylight savings time
) ; eval-when

(fli:define-foreign-function (libc-mktime "mktime")
    ((ptime (:pointer libc-time)))
  :module (load-time-value *c-rtl-module*)
  :result-type time-t
  :calling-convention :cdecl)

(fli:define-foreign-function (libc-gmtime "gmtime")
    ((unix-time (:reference-pass time-t)))
  :module (load-time-value *c-rtl-module*)
  :result-type (:pointer libc-time)
  :calling-convention :cdecl)

;; Misfunctions on Windows NT? (always return the same as gmtime: tm-isdst = 0)
(fli:define-foreign-function (libc-localtime "localtime")
    ((unix-time (:reference-pass time-t)))
  :module (load-time-value *c-rtl-module*)
  :result-type (:pointer libc-time)
  :calling-convention :cdecl)

(defun unix-mktime (year month day &optional (hour 0) (min 0) (sec 0))
 ;; Interpret arguments as a local time parts.
 ;; Value: Unix time.
 ;; Set the tm_isdst field to 0 to indicate that standard time is in effect, 
 ;; or to a value greater than 0 to indicate that daylight savings time is in effect,
 ;; or to a value less than zero to have the C run-time library code compute 
 ;; whether standard time or daylight savings time is in effect. 
 ;; Surprise: Works correctly! Assumes that input parameters correspond to Windows
 ;;           regional settings of timezone and DST (USA DST rules?)
 ;;
  (declare (type integer year month day hour min sec))
  (fli:with-dynamic-foreign-objects ((libc-time libc-time))
    (setf (fli:foreign-slot-value libc-time 'tm-year) year)
    (setf (fli:foreign-slot-value libc-time 'tm-mon)  month)
    (setf (fli:foreign-slot-value libc-time 'tm-mday) day)
    (setf (fli:foreign-slot-value libc-time 'tm-hour) hour)
    (setf (fli:foreign-slot-value libc-time 'tm-min)  min)
    (setf (fli:foreign-slot-value libc-time 'tm-sec)  sec)
    (setf (fli:foreign-slot-value libc-time 'tm-yday) 0)
    (setf (fli:foreign-slot-value libc-time 'tm-wday) 0)
    (setf (fli:foreign-slot-value libc-time 'tm-isdst) -1) ; let compute itself
    (libc-mktime libc-time)))

(defun unix-gmtime (unix-time)
  (declare (type integer unix-time))
  (let ((libc-time (libc-gmtime unix-time))) 
    (values (fli:foreign-slot-value libc-time 'tm-sec) 	 ; seconds
            (fli:foreign-slot-value libc-time 'tm-min)	 ; minutes
            (fli:foreign-slot-value libc-time 'tm-hour)	 ; hours
            (fli:foreign-slot-value libc-time 'tm-mday)	 ; day of month
            (fli:foreign-slot-value libc-time 'tm-mon)	 ; month
            (fli:foreign-slot-value libc-time 'tm-year)	 ; year
            (fli:foreign-slot-value libc-time 'tm-wday)	 ; day of week
            (fli:foreign-slot-value libc-time 'tm-yday)	 ; day of year
            (fli:foreign-slot-value libc-time 'tm-isdst)))) ; daylight savings time = 0

(defun unix-localtime (unix-time)
  (declare (type integer unix-time))
  (let ((libc-time (libc-localtime unix-time)))
    (values (fli:foreign-slot-value libc-time 'tm-sec) 	 ; seconds
            (fli:foreign-slot-value libc-time 'tm-min)	 ; minutes
            (fli:foreign-slot-value libc-time 'tm-hour)	 ; hours
            (fli:foreign-slot-value libc-time 'tm-mday)	 ; day of month
            (fli:foreign-slot-value libc-time 'tm-mon)	 ; month
            (fli:foreign-slot-value libc-time 'tm-year)	 ; year
            (fli:foreign-slot-value libc-time 'tm-wday)	 ; day of week
            (fli:foreign-slot-value libc-time 'tm-yday)	 ; day of year
            (fli:foreign-slot-value libc-time 'tm-isdst)))) ; daylight savings time

;;; Timezone initialization - sets time environment variables of C RTL
(fli:define-foreign-function (libc-tzset "_tzset") ()
  :module (load-time-value *c-rtl-module*)
  :result-type :void
  :calling-convention :cdecl)

(eval-when (:load-toplevel :execute) (libc-tzset))	; to initialize library

;; Current timezone values:
;; - difference in seconds between coordinated universal time and local time
;; - daylight saving
(fli:define-foreign-variable (libc-timezone "_timezone") :type time-t)
(fli:define-foreign-variable (libc-daylight "_daylight") :type :int)
;;(fli:define-foreign-variable (libc-tzname "_tzname")
;;	:type (:c-array (:pointer :char) 2))
;; (fli:convert-from-foreign-string (fli:foreign-aref tzname 0))
;; => "Russian Standard Time"
;; (fli:convert-from-foreign-string (fli:foreign-aref tzname 1))
;; => "Russian Daylight Time"

(defun unix-timezone (&optional unix-time)
 "Returns information about this timezone relative to moment of UNIX-TIME.
 Argumens: If UNIX-TIME is null, assume the current time.
 Values: This zone displacement as seconds East of UCT,
  and boolean daylight-saving-p."
  (if unix-time
      ;; Beware of reusing the same C structure when returning results for both calls
      (let* ((localtime (libc-localtime unix-time))
             (tm-sec (fli:foreign-slot-value localtime 'tm-sec))
             (tm-min (fli:foreign-slot-value localtime 'tm-min))
             (tm-year (fli:foreign-slot-value localtime 'tm-year))
             (tm-hour (fli:foreign-slot-value localtime 'tm-hour))
             (tm-yday (fli:foreign-slot-value localtime 'tm-yday))
             (tm-isdst (fli:foreign-slot-value localtime 'tm-isdst))
             (gmtime (libc-gmtime unix-time))
             (year (- tm-year (fli:foreign-slot-value gmtime 'tm-year))))
        (declare (type fixnum tm-year tm-hour tm-yday tm-isdst year))
        (values (+ (- tm-sec (fli:foreign-slot-value gmtime 'tm-sec))
                   (* 60
                      (+ (- tm-min (fli:foreign-slot-value gmtime 'tm-min))
                         (* 60
                            (+ (- tm-hour (fli:foreign-slot-value gmtime 'tm-hour))
                               (* 24
                                  (if (zerop year)		; tm-year's are the same
                                      (- tm-yday
                                         (fli:foreign-slot-value gmtime 'tm-yday))
                                      (signum year))))))))
                (plusp tm-isdst)))
                ;(case tm-isdst (1 t) (0 nil) (otherwise :unknown))))
      ;; else calculate offset for the current time
      (let ((tm-isdst (/= (libc-daylight) 0)))
        (values (- (if tm-isdst 3600 0) (libc-timezone))
                tm-isdst))))

#|(fli:define-foreign-function (msvc-time "time")
    ((ptime (:reference-pass time-t :allow-null t)))
  :module (load-time-value *c-rtl-module*)
  :result-type time-t
  :calling-convention :cdecl)|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Win32 hacking
;;; In Windows timezone setting, the Caucasus Standard Time can be set for Moscow
;;; without auto adjusting clock.

(eval-when (:compile-toplevel :load-toplevel :execute)

 (fli:define-c-struct |SYSTEMTIME| 
  (wYear win32:WORD)           ; year
  (wMonth win32:WORD)          ; month; January = 1, February = 2, and so on
  (wDayOfWeek win32:WORD)      ; day of the week; Sunday = 0, Monday = 1, and so on
  (wDay win32:WORD)            ; day of the month
  (wHour win32:WORD)           ; hour
  (wMinute win32:WORD)         ; minute
  (wSecond win32:WORD)         ; second
  (wMilliseconds win32:WORD))  ; millisecond

(fli:define-c-struct |FILETIME|
  (low win32:DWORD)			; DWORD dwLowDateTime
  (high win32:DWORD))			; DWORD dwHighDateTime

 (fli:define-c-struct |TIME_ZONE_INFORMATION| 
  (Bias :long)                       ; current bias, minutes: UTC = local time + bias (Moscow -180)
  (StandardName (:c-array :wchar-t 32)) ; Unicode string ("Russian Standard Time")
  (StandardDate |SYSTEMTIME|)        ; wMonth=10, wDayOfWeek=0, wDay=5 (last Sun), wHour=3
  (StandardBias :long)               ; Usually 0
  (DaylightName (:c-array :wchar-t 32)) ; Unicode string ("Russian Daylight Time")
  (DaylightDate |SYSTEMTIME|)        ; wMonth=3, wDayOfWeek=0, wDay=5 (last Sun), wHour=2
  (DaylightBias :long))              ; bias value to be used during dst (usually -60)

 (defconstant TIME_ZONE_ID_INVALID #xFFFFFFFF)
 (defconstant TIME_ZONE_ID_UNKNOWN 0) ; no transition dates specified
 (defconstant TIME_ZONE_ID_STANDARD 1)
 (defconstant TIME_ZONE_ID_DAYLIGHT 2)
) ; eval-when

(define-win32api |GetSystemTime| ((lpSystemTime (:pointer |SYSTEMTIME|))) :void)
(define-win32api |GetLocalTime| ((lpSystemTime (:pointer |SYSTEMTIME|))) :void)
(define-win32api |GetTimeZoneInformation|
  ((lpTimeZoneInformation (:pointer |TIME_ZONE_INFORMATION|)))
  win32:DWORD)
(define-win32api |SystemTimeToTzSpecificLocalTime|
  ((lpTimeZone (:reference-pass |TIME_ZONE_INFORMATION| :allow-null t))
                   ; if lpTimeZone is NULL, uses the currently active time zone
   (lpSystemTime (:pointer |SYSTEMTIME|))
   (lpLocalTime (:pointer |SYSTEMTIME|)))
  :boolean)
(define-win32api |FileTimeToSystemTime|
  ((lpFileTime (:pointer |FILETIME|))		; IN CONST FILETIME *lpFileTime
   (lpSystemTime (:pointer |SYSTEMTIME|)))	; OUT LPSYSTEMTIME lpSystemTime
  :boolean)


(defun decode-win32-system-time (system-time)
  (values (fli:foreign-slot-value system-time 'wYear)
          (fli:foreign-slot-value system-time 'wMonth)	
          (fli:foreign-slot-value system-time 'wDay)
          (fli:foreign-slot-value system-time 'wHour)	
          (fli:foreign-slot-value system-time 'wMinute)
          (fli:foreign-slot-value system-time 'wSecond)
          (fli:foreign-slot-value system-time 'wMilliseconds)
          (fli:foreign-slot-value system-time 'wDayOfWeek)))

(defun get-win32-system-time ()
  "Return current UTC time"
  (fli:with-dynamic-foreign-objects ((system-time |SYSTEMTIME|))
    (|GetSystemTime| system-time)
    (decode-win32-system-time system-time)))

(defun get-win32-local-time ()
  "Return current UTC time"
  (fli:with-dynamic-foreign-objects ((system-time |SYSTEMTIME|))
    (|GetLocalTime| system-time)
    (decode-win32-system-time system-time)))

(defun get-win32-timezone-information ()
  (fli:with-dynamic-foreign-objects ((tzi |TIME_ZONE_INFORMATION|))
    (let ((result (|GetTimeZoneInformation| tzi)))
      (case result
        (#.TIME_ZONE_ID_UNKNOWN
         (values result
                 (fli:foreign-slot-value tzi 'Bias)
                 (fli:foreign-slot-value tzi 'StandardBias)
                 (fli:foreign-slot-value tzi 'DaylightBias)
                 nil nil		; no transition dates used for this timezone
                 (fli:convert-from-foreign-string
                  (fli:foreign-slot-pointer tzi 'StandardName)
                  :external-format :unicode) ;yl:*internal-format*
                 nil))
        ((#.TIME_ZONE_ID_STANDARD #.TIME_ZONE_ID_DAYLIGHT)
         (values result
                 (fli:foreign-slot-value tzi 'Bias)
                 (fli:foreign-slot-value tzi 'StandardBias)
                 (fli:foreign-slot-value tzi 'DaylightBias)
                 (multiple-value-list (decode-win32-system-time
                                       (fli:foreign-slot-value tzi 'StandardDate
                                             :copy-foreign-object nil)))
                 (multiple-value-list (decode-win32-system-time
                                       (fli:foreign-slot-value tzi 'DaylightDate
                                             :copy-foreign-object nil)))
                 ;; Signals error on the Russian version of Windows: 
                 ;;  External format FLI:ASCII-WCHAR got error reading
                 ;;  #<Pointer to type ... = #x007CBCEC> at position 0:
                 ;;  Non-ASCII code ...
                 ;(fli:foreign-slot-value tzi 'StandardName :type :ef-wc-string
                 ;                                          :copy-foreign-object t)
                 ;(fli:foreign-slot-value tzi 'DaylightName :type :ef-wc-string
                 ;                                          :copy-foreign-object t)
	         ;; the pointer does not need to be of the correct type
                 (fli:convert-from-foreign-string
                  (fli:foreign-slot-pointer tzi 'StandardName)
                  :external-format :unicode) ;yl:*internal-format*
                 (fli:convert-from-foreign-string
                  (fli:foreign-slot-pointer tzi 'DaylightName)
                  :external-format :unicode) ;yl:*internal-format*
                 ))
        (otherwise (values))))))

(defun win32-timezone ()
 ;;; Values: current timezone displacement in seconds and dst-p
  (fli:with-dynamic-foreign-objects ((tzi |TIME_ZONE_INFORMATION|))
    (case (|GetTimeZoneInformation| tzi)
      ((#.TIME_ZONE_ID_UNKNOWN #.TIME_ZONE_ID_STANDARD)
       (values (* (+ (fli:foreign-slot-value tzi 'Bias)
                     (fli:foreign-slot-value tzi 'StandardBias))
                  -60)
               nil))
      (#.TIME_ZONE_ID_DAYLIGHT
       (values (* (+ (fli:foreign-slot-value tzi 'Bias)
                     (fli:foreign-slot-value tzi 'DaylightBias))
                  -60)
               t)))))

