;;; -*- Mode: Lisp -*-
;;; Local-Time - Predicates and operations on time instances
;;; Copyright (c) 2002-2006 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Implementation Note: Coding via loop leads to increasing size compiled code the twice.

(in-package :ystok.local-time)

;;; Arithmetic predicates

(defmacro set-normalized-lt-day-sec ((var-day day) (var-sec sec))
  (with-gensyms (sec-days)
    `(let ((,sec-days 0))
       (declare (fixnum ,sec-days))
       (multiple-value-setq (,sec-days ,var-sec) (floor ,sec +secs-per-day+))
       (setq ,var-day (the fixnum (+ ,day (the fixnum ,sec-days)))))))

(defmacro with-relative-components ((x-day x-sec x-msec y-day y-sec y-msec) x y
                                   both-dates-form &body body)
 ;;; If one of the args x or y is utc-time but the other is not, convert the latter
  ;; to UTC zone while defaulting dates by zero sec and msec.
  ;; Otherwise extract them as is while defaulting dates by zero sec and msec.
  ;; NB: We default the date sec component by zero alike computing its timezone.
  ;; NB: Is it correct ignoring completely zone difference when
  ;;     both are dates or local-times and *session-timezone* = t?
 `(let ((,x-day (the fixnum (local-time-day ,x))) (,x-sec 0) (,x-msec 0)
        (,y-day (the fixnum (local-time-day ,y))) (,y-sec 0) (,y-msec 0))
    (declare #-debug (optimize (speed 3) (safety 0) (debug 0)
                               #+lispworks (hcl:fixnum-safety 0))
             (fixnum ,x-day ,x-sec ,x-msec ,y-day ,y-sec ,y-msec))
    (cond ((utc-time-p ,x)
           (setq ,x-sec (the fixnum (local-time-sec ,x))
                 ,x-msec (the fixnum (local-time-msec ,x)))
           (cond ((utc-time-p ,y)
                  (setq ,y-sec (the fixnum (local-time-sec ,y))
                        ,y-msec (the fixnum (local-time-msec ,y))))
                 ((local-time-p ,y)
                  (set-normalized-lt-day-sec (,y-day ,y-day)
                      (,y-sec (the fixnum (- (the fixnum (local-time-sec ,y))
                                             (the fixnum (values (timezone ,y)))))))
                  (setq ,y-msec (the fixnum (local-time-msec ,y))))
                 ((set-normalized-lt-day-sec (,y-day ,y-day)
                      (,y-sec (the fixnum (- (the fixnum (values (timezone ,y))))))))))
          ((local-time-p ,x)
           (cond ((utc-time-p ,y)
                  (set-normalized-lt-day-sec (,x-day ,x-day)
                      (,x-sec (the fixnum (- (the fixnum (local-time-sec ,x))
                                             (the fixnum (values (timezone ,x)))))))
                  (setq ,x-msec (the fixnum (local-time-msec ,x))
                        ,y-sec (the fixnum (local-time-sec ,y))
                        ,y-msec (the fixnum (local-time-msec ,y))))
                 ((local-time-p ,y)
                  (setq ,x-sec (the fixnum (local-time-sec ,x))
                        ,x-msec (the fixnum (local-time-msec ,x))
                        ,y-sec (the fixnum (local-time-sec ,y))
                        ,y-msec (the fixnum (local-time-msec ,y)))) ))
          ;; (assert (date-p ,x))
          ((utc-time-p ,y)
           (set-normalized-lt-day-sec (x-day x-day)
                      (,x-sec (the fixnum (- (the fixnum (values (timezone ,x)))))))
           (setq ,y-sec (the fixnum (local-time-sec ,y))
                 ,y-msec (the fixnum (local-time-msec ,y))))
          ((local-time-p ,y)
           (setq ,y-sec (the fixnum (local-time-sec ,y))
                 ,y-msec (the fixnum (local-time-msec ,y))))
          ,@(when both-dates-form `((,both-dates-form))) )
    ,@body))

(defmethod local-time=2 ((x date) (y date))
  (with-relative-components (x-day x-sec x-msec y-day y-sec y-msec) x y
      (return-from local-time=2 (= x-day y-day))
    (and (= x-sec y-sec)
         (= x-day y-day)
         (= x-msec y-msec))))

(defun local-time= (&rest args)
  (do* ((args args rest)
        (rest (rest args) (rest args)))
       ((null rest) t)
    (unless (local-time=2 (first args) (first rest)) (return nil))))
  ;(loop for (x y) on args while y always (local-time=2 x y)))

(define-compiler-macro local-time= (&whole whole &rest args)
  (if (= (length whole) 3)
      `(local-time=2 ,(first args) ,(second args))
      whole))

(defun local-time/= (&rest args)
  (do* ((args args rest)
        (x (first args) (first args))
        (rest (rest args) (rest args)))
       ((null rest) t)
    (dolist (y rest)
      (when (local-time=2 x y) (return-from local-time/= nil)))))
  ;(loop for (x . rest) on args while rest
  ;      do (dolist (y rest) (when (local-time=2 x y) (return-from local-time/= nil)))
  ;      finally (return t)))

(define-compiler-macro local-time/= (&whole whole &rest args)
  (if (= (length whole) 3)
      `(not (local-time=2 ,(first args) ,(second args)))
      whole))

(defmethod local-time<2 ((x date) (y date))
  (with-relative-components (x-day x-sec x-msec y-day y-sec y-msec) x y
      (return-from local-time<2 (< x-day y-day))
    (or (< x-day y-day)
        (and (= x-day y-day)
             (or (< x-sec y-sec)
                 (and (= x-sec y-sec)
                      (< x-msec y-msec)))))))

(defun local-time< (&rest args)
  (declare (optimize (speed 3) (safety 1) (space 0) (debug 0) (compilation-speed 0)))
  (do* ((args args rest)
        (rest (rest args) (rest args)))
       ((null rest) t)
    (unless (local-time<2 (first args) (first rest))
      (return nil))))
  ;(loop for (x y) on args while y always (local-time<2 x y))

(define-compiler-macro local-time< (&whole whole &rest args)
  (if (and (eq (first whole) 'local-time<) (= (length whole) 3))
      `(local-time<2 ,(first args) ,(second args))
      whole))

(defun local-time<= (&rest args)
  (do* ((args args rest)
        (rest (rest args) (rest args)))
       ((null rest) t)
    (when (local-time<2 (first rest) (first args))
      (return nil))))
  ;(loop for (x y) on args while y never (local-time<2 y x)))

(define-compiler-macro local-time<= (&whole whole &rest args)
  (if (= (length whole) 3)
      `(not (local-time<2 ,(second args) ,(first args)))
      whole))

(defun local-time>2 (x y) (local-time<2 y x))

(defun local-time> (&rest args)
  (do* ((args args rest)
        (rest (rest args) (rest args)))
       ((null rest) t)
    (unless (local-time<2 (first rest) (first args))
      (return nil))))
  ;(loop for (x y) on args while y always (local-time<2 y x)))

(define-compiler-macro local-time> (&whole whole &rest args)
  (if (= (length whole) 3)
      `(local-time<2 ,(second args) ,(first args))
      whole))

(defun local-time>= (&rest args)
  (do* ((args args rest)
        (rest (rest args) (rest args)))
       ((null rest) t)
    (when (local-time<2 (first args) (first rest))
      (return nil))))
  ;(loop for (x y) on args while y never (local-time<2 x y)))

(define-compiler-macro local-time>= (&whole whole &rest args)
  (if (= (length whole) 3)
      `(not (local-time<2 ,(first args) ,(second args)))
      whole))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Arithmetic operators
;;; Valid operators involving datetimes and intervals
;;;   (Operator	 Operand-1	Operand-2)  =>	Result Type
;;;	-	 datetime	datetime	duration
;;;	+ or -	 datetime	duration	datetime
;;;	+	 duration	datetime	datetime
;;;	+ or -	 duration	duration	duration
;;;	* or /	 duration	number		duration
;;;	*	 number		duration	duration
;;;	/	 duration	duration	number	 (absent in SQL-99)

;;; Binary plus
(defmethod local-time+2 ((x local-time) (y ym-duration))
 ;;; Value: If x is the last day of the month or if the resulting month has fewer days
  ;; than the day component of x, then the result is the last day of the resulting 
  ;; month. Otherwise, the result has the same day component as x (Oracle's ADD_MONTHS).
  ;; NB: If the resulting year <= 0, then works incorrectly
  (multiple-value-bind (ms seconds minutes hours day month year) (decode-as-local x)
    (declare (fixnum year month day))
    (multiple-value-bind (result-year m)
        (truncate (+ (* year 12) month (1- (duration-month y))) 12)
      (declare (fixnum result-year m))
      (let* ((result-month (1+ m))
             (days-in-result-month (days-in-month result-year result-month)))
        ;; year and month are of the resulting time
        (encode-local-time ms seconds minutes hours
                           (if (or (> day days-in-result-month)	
                                   (= day (days-in-month year month)))	; last day
                               days-in-result-month
                               day)
                           result-month result-year
                           *session-timezone* (type-of x) nil) ))))

(defmethod local-time+2 ((x date) (y ym-duration))
  (multiple-value-bind (year month day) (decode-date x)
    (declare (fixnum year month day))
    (multiple-value-bind (result-year m)
        (truncate (+ (* year 12) month (1- (duration-month y))) 12)
      (declare (fixnum result-year m))
      (let* ((result-month (1+ m))
             (days-in-result-month (days-in-month result-year result-month)))
        (declare (fixnum result-month days-in-result-month))
        ;; year and month are of the resulting time
        (encode-date result-year result-month
                     (if (or (> day days-in-result-month)	
                             (= day (days-in-month year month)))	; last day
                         days-in-result-month
                         day))))))

(defmethod local-time+2 ((x ym-duration) (y date))
  (local-time+2 y x))

(defmethod local-time+2 ((x local-time) (y dt-duration))
  (make-local-time :day (the fixnum (+ (local-time-day x) (duration-day y)))
                   :sec (the fixnum (+ (local-time-sec x) (duration-sec y)))
                   :msec (the fixnum (+ (local-time-msec x) (duration-msec y)))
                   :zone (if (utc-time-p x) (local-time-zone x) nil)))

(defmethod local-time+2 ((x dt-duration) (y local-time))
  (make-local-time :day (the fixnum (+ (duration-day x) (local-time-day y)))
                   :sec (the fixnum (+ (duration-sec x) (local-time-sec y)))
                   :msec (the fixnum (+ (duration-msec x) (local-time-msec y)))
                   :zone (if (utc-time-p y) (local-time-zone y) nil)))

(defmethod local-time+2 ((x local-time) (y number))
  (multiple-value-bind (y-day y-sec y-msec) (%number-day-sec-msec y)
    (declare (fixnum y-day y-sec y-msec))
    (make-local-time :day (the fixnum (+ (local-time-day x) y-day))
                     :sec (the fixnum (+ (local-time-sec x) y-sec))
                     :msec (the fixnum (+ (local-time-msec x) y-msec))
                     :zone (if (utc-time-p x) (local-time-zone x) nil))))

(defmethod local-time+2 ((x number) (y local-time))
  (multiple-value-bind (x-day x-sec x-msec) (%number-day-sec-msec x)
    (declare (fixnum x-day x-sec x-msec))
    (make-local-time :day (the fixnum (+ x-day (local-time-day y)))
                     :sec (the fixnum (+ x-sec (local-time-sec y)))
                     :msec (the fixnum (+ x-msec (local-time-msec y)))
                     :zone (if (utc-time-p y) (local-time-zone y) nil))))

(defmethod local-time+2 ((x date) (y dt-duration))
  (with-dt (y-day y-sec y-msec) y
    (let ((day (the fixnum (+ (local-time-day x) y-day))))
      (declare (fixnum day))
      (cond ((= y-sec y-msec 0)
             (make-date :day day))
            ((or (neq *default-encode-type* 'utc-time) (null *session-timezone*))
             (make-local-time :day day  :sec y-sec  :msec y-msec))
            ((let ((timezone (if (eq *session-timezone* t)
                                 (%lt-timezone day y-sec)
                                 *session-timezone*)))
               (make-local-time :day day
                                :sec (- y-sec timezone)
                                :msec y-msec  
                                :zone timezone)))))))

(defmethod local-time+2 ((x dt-duration) (y date))
  (local-time+2 y x))

(defmethod local-time+2 ((x date) (y number))
  (multiple-value-bind (y-day y-sec y-msec) (%number-day-sec-msec y)
    (declare (fixnum y-day y-sec y-msec))
    (let ((day (the fixnum (+ (local-time-day x) y-day))))
      (declare (fixnum day))
      (cond ((= y-sec y-msec 0)
             (make-date :day day))
            ((or (neq *default-encode-type* 'utc-time) (null *session-timezone*))
             (make-local-time :day day  :sec y-sec  :msec y-msec))
            ((let ((timezone (if (eq *session-timezone* t)
                                 (%lt-timezone day y-sec)
                                 *session-timezone*)))
               (make-local-time :day day  
                                :sec (- y-sec timezone)
                                :msec y-msec  
                                :zone timezone)))))))

(defmethod local-time+2 ((x number) (y date))
  (local-time+2 y x))

(defun local-time+ (arg &rest args)
 "Add durations and datetimes. Return a new datetime or duration instance."
  (reduce #'local-time+2 args :initial-value arg))

;;; Binary minus
(defmethod local-time-2 ((x date) (y date))
  (with-relative-components (x-day x-sec x-msec y-day y-sec y-msec) x y nil
    (make-dt-duration :day (the fixnum (- x-day y-day))
                      :sec (the fixnum (- x-sec y-sec))
                      :msec (the fixnum (- x-msec y-msec)))))

(defmethod local-time-2 ((x local-time) (y dt-duration))
  (make-local-time :day (the fixnum (- (local-time-day x) (duration-day y)))
                   :sec (the fixnum (- (local-time-sec x) (duration-sec y)))
                   :msec (the fixnum (- (local-time-msec x) (duration-msec y)))
                   :zone (if (utc-time-p x) (local-time-zone x) nil)))

(defmethod local-time-2 ((x local-time) (y number))
  (multiple-value-bind (y-day y-sec y-msec) (%number-day-sec-msec y)
    (declare (fixnum y-day y-sec y-msec))
    (make-local-time :day (the fixnum (- (local-time-day x) y-day))
                     :sec (the fixnum (- (local-time-sec x) y-sec))
                     :msec (the fixnum (- (local-time-msec x) y-msec))
                     :zone (if (utc-time-p x) (local-time-zone x) nil))))

(defmethod local-time-2 ((x date) duration)
  (local-time+2 x (duration- duration)))

(defun local-time- (arg &rest args)
  "Subtract durations from a datetime or durations.
 Return a new datetime or duration value."
  (if args
      (reduce #'local-time-2 args :initial-value arg)
      (duration- arg)))


#||;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; EVALUATION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun local-time/ (x y)
 ;;; Args: x, y  Both are durations
)

(defmethod local-time+22 ((x local-time) (y ym-duration))
 ;;; If x is the last day of the month or if the resulting month has fewer days
  ;; than the day component of x, then the result is the last day of the resulting
  ;; month. Otherwise, the result has the same day component as x.
  ;; NB: a la Oracle's ADD_MONTHS.
  (let ((months (duration-month y))
        (x-day (local-time-day x)))
    (multiple-value-bind (year month day) (%lt-day-ymd x-day)
      (declare (fixnum year month day))
      (let* ((days-in-begin-month (days-in-month year month))
             (days (cond ((> months 0)
                          (loop repeat months
                                sum (days-in-month year month)
                                if (= month 12) do (setq month 1  year (1+ year))
                                else do (incf month)))
                         ((< months 0)
                          (- (loop repeat (- months)
                                   if (= month 1) do (setq month 12  year (1- year))
                                   else do (decf month)
                                   sum (days-in-month year month))))
                         (0)))
             (days-in-end-month (days-in-month year month)))
        ;; year and month are resulting
        (make-local-time :day (+ x-day days
                                 (if (or (= day days-in-begin-month)	; last day
                                         (> day days-in-end-month))
                                     (- days-in-end-month day)
                                     0))
                         :sec (local-time-sec x)
                         :msec (local-time-msec x)
                         :zone (if (utc-time-p x) (local-time-zone x) nil))
) ) ) )
||#
