;;; -*- Mode: Lisp -*-
;;; Local-Time - High-level convenience functions
;;; Copyright (c) 2002-2009 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; We always regards *session-timezone*

(in-package :ystok.local-time)

;;; Simplified date equality and coerce

(defun date= (x y)
  (= (local-time-day x) (local-time-day y)))

(defun date-local-time (arg)
 ;;; Coerce date or date interval to local-time or interval of local-time instances
  (cond ((local-time-p arg)
         arg)
        ((date-p arg)
         (%make-local-time :day (local-time-day arg)))	; simply set other slots to 0
        #+ys-product					; CAUTION: compiled to the same bin!
        ((interval-p arg)
         (let ((start (interval-start arg))
               (end (interval-end arg)))
           (let ((new-start (date-local-time start))
                 (new-end (date-local-time end)))
             (if (and (eq new-start start) (eq new-end end))
                 arg
                 (make-interval :start new-start :end new-end)))))))

(defun local-time-date (local-time)
  (cond #|((utc-time-p local-time)
         (multiple-value-bind (msec seconds minutes hours)
             (decode-as-local local-time nil)
           (declare (ignore msec))
           (values hours minutes seconds)))|#
        ((local-time-p local-time)
         (make-date :day (local-time-day local-time)))
        ((date-p local-time)
         local-time)))

(defun local-time-hms (local-time)
 "Return as multiple values the hour, minute and second"
  (cond ((utc-time-p local-time)
         (multiple-value-bind (msec seconds minutes hours)
             (decode-as-local local-time nil)
           (declare (ignore msec))
           (values hours minutes seconds)))
        ((local-time-p local-time)
         (%lt-sec-hms (local-time-sec local-time)))
        ((date-p local-time)
         (values 0 0 0))))

(defun merge-date-time (date time)
 ;;; Args: time can be an interval, date cannot!
  ;; Value: Instance of local-time.
  ;; CAUTION: We must not return an instance of date when tim represents midnight!
  (cond ((local-time-p date)
         date)
        #+ys-product
        ((interval-p time)
         (make-interval :start (merge-date-time date (interval-start time))
                        :end (merge-date-time date (interval-end time))))
        ((integerp time)				; tim in minutes
         (multiple-value-bind (day minutes) (truncate time #.(* 60 24))
           (declare (fixnum day minutes))
           (%make-local-time :day (+ day (the fixnum (local-time-day date)))
                             :sec (* minutes 60))))
         ;(local-time+ date (/ time #.(* 60 24))))
        ((utc-time-p time)
         (%make-utc-time :day (local-time-day date)
                         :sec (local-time-sec time)
                         :msec (local-time-msec time)
                         :zone (local-time-zone time)))
        (t (%make-local-time :day (local-time-day date)
                             :sec (local-time-sec time)
                             :msec (local-time-msec time)))))

(defun week-day-number (date)
 "Return ISO8601 ordinal day number in the week: 1=Mon,...,7=Sun"
  (if (utc-time-p date)
      (nth-value 7 (decode-as-local date nil))
      (1+ (%lt-day-dow (local-time-day date)))))

(defun week-day-name (date)
  (svref yl:*weekday-names* (1- (week-day-number date))))

(defun week-day-abbrev (date)
  (svref yl:*weekday-abbrevs* (1- (week-day-number date))))

(defun week (date &optional (iso t))
 ;;; Return calendar week number:
  ;; Args: iso = True: 1,...,53.
  ;;		 The first week is assumed to have at least four days in the year.
  ;;	       = False: 1,...,54 (54 for leap years only, where December 31 is Monday).
  ;;		 The first calendar week is the week in which January 1 occurs.
  ;; Values: 1) calendar week number
  ;;	     2) year (actual if iso = nil)
  ;; NB: ISO   .
  ;;	             # .
  ;; Oracle NB:
  ;;	In addition to ISO concept ("IW" format specifier), there two features:
  ;;	W   Week of month (1-5) where week 1 starts on the first day of the month
  ;;	    and ends on the seventh.
  ;;	WW  Week of year (1-53) where week 1 starts on the first day of the year
  ;;	    and continues to the seventh day of the year.
  (if iso
      (if (utc-time-p date)
          (with-adjusted-hms (hour-days hours minutes seconds)
              (+ (local-time-sec date) (local-time-zone date))
            (%lt-day-week-iso (+ (local-time-day date) hour-days)))
          (%lt-day-week-iso (local-time-day date)))
      (if (utc-time-p date)
          (with-adjusted-hms (hour-days hours minutes seconds)
              (+ (local-time-sec date) (local-time-zone date))
            (%lt-day-week (+ (local-time-day date) hour-days)))
          (%lt-day-week (local-time-day date)))))

(defun current-timestamp (&optional precision (encode-type *default-encode-type*))
 "Return the current time as a local-time instance"
 ;;; Args: precision   Decimal digits in the seconds part; possible values: 
  ;;		       3 (default for Windows) or
  ;;                   0 (always on other platforms - cannot determine millisecs for now).
  ;;       encode-type Either local-time (corresponds to the SQL LOCALTIMESTAMP)
  ;;                   or utc-time (corresponds to the SQL CURRENT_TIMESTAMP).
  ;; NB: Cannot determine milliseconds on all platforms for now
  #+(and lispworks win32)
  (multiple-value-bind (year month day hours minutes seconds msec)
      (get-win32-local-time) 						; don't check
    (encode-local-time (if (and precision (zerop precision)) 0 msec)
                       seconds minutes hours day month year
                       *session-timezone* encode-type nil))
  #-(and lispworks win32)
  (let ((zone (if (numberp *session-timezone*) *session-timezone* (unix-timezone))))
    (if (eq encode-type 'utc-time)
        (local-time :universal (get-universal-time) :zone zone)
        (local-time :universal (+ (get-universal-time) zone) :zone nil))))

(declaim (inline localtimestamp))
(defun localtimestamp (&optional precision) (current-timestamp precision 'local-time))

(defun current-date (&optional (encode-type 'date))
 "Return the current time as a local-time instance"
  #+(and lispworks win32)
  (multiple-value-bind (year month day) (get-win32-local-time)  	; don't check
    (declare (fixnum year month day))
    (if (eq encode-type 'date)
        (encode-date year month day nil)
        (encode-local-time 0 0 0 0 day month year *session-timezone* encode-type nil)))
               ;(if (numberp *session-timezone*) *session-timezone* (win32-timezone))
  #-(and lispworks win32)
  (multiple-value-bind (seconds minutes hours day month year) (get-decoded-time)
    (declare (ignore seconds minutes hours))	 			; don't check
    (if (eq encode-type 'date)
        (encode-date year month day nil)
        (encode-local-time 0 0 0 0 day month year *session-timezone* encode-type nil))))
                       ;(if (numberp *session-timezone*) *session-timezone*
                       ;    (* (+ zone (if dst-p -1 0)) -3600))

(defun current-year ()
  #+(and lispworks win32) (values (get-win32-local-time))
  #-(and lispworks win32) (nth-value 5 (get-decoded-time)))


(defgeneric time-component (arg component &optional decode-type))
(defmethod time-component ((arg local-time) component 
                           &optional (decode-type *default-encode-type*))
  (multiple-value-bind (msec second minute hour day month year week-day-number)
      (if (eq decode-type 'local-time)
          (decode-as-local arg)
          (decode-as-utc arg))
    (ecase component
      (:millisecond	msec)
      (:second		(if (zerop msec) second (+ second (* msec 0.001)))) ; ?
      (:minute		minute)
      (:hour		hour)
      (:day		day)
      (:month		month)
      (:year		year)
      (:week-day-number	week-day-number)
      (:week-day-keyword (svref *weekday-keywords* (1- week-day-number)))
      (:timezone-hour	0)
      (:timezone-minute	0)
      (:week		(if (and (utc-time-p arg) (eq decode-type 'local-time))
                            (week arg)				  ; adjustment needed
                            (%lt-day-week (local-time-day arg)))) ; straightforward
) ) )

(defmethod time-component ((arg date) component &optional decode-type)
  (declare (ignore decode-type))
  (ecase component
    (:day		(nth-value 2 (decode-date arg)))
    (:month		(nth-value 1 (decode-date arg)))
    (:year		(values (decode-date arg)))
    (:week-day-number	(1+ (%lt-day-dow (local-time-day arg))))
    (:week-day-keyword	(svref *weekday-keywords* (%lt-day-dow (local-time-day arg))))
    (:week		(%lt-day-week (local-time-day arg)))
) )

(defmethod time-component ((arg ym-duration) component &optional decode-type)
  (declare (ignore decode-type))
  (multiple-value-bind (year month) (decode-duration arg)
    (ecase component
      (:month month)
      (:year  year))))

(defmethod time-component ((arg dt-duration) component &optional decode-type)
  (declare (ignore decode-type))
  (multiple-value-bind (day hour minute second msec) (decode-duration arg)
    (ecase component
      (:millisecond	msec)
      (:second		(if (zerop msec) second (+ second (* msec 0.001)))) ; ?
      (:minute		minute)
      (:hour		hour)
      (:day		day)
) ) )


;;; keep-symbols?
;;; Arg: date ::= lt:date or integer (means local-time-day component - used by calendar).
;;; The finish-of-... family return the last date within the interval assumed.

(defun start-of-week (date)
  (local-time- date (1- (week-day-number date))))
(defun end-of-week (date)
  (local-time+ date (- 8 (week-day-number date))))
(defun finish-of-week (date)
  (local-time- (end-of-week date) 1))
  ;(local-time+ date (- 7 (week-day-number date))))

(defun start-of-10days (date)
  (multiple-value-bind (year month day) (decode-date date)
    (encode-date year month (cond ((<= 1 day 10) 1)
                                  ((<= 11 day 20) 11)
                                  (t 21)) nil)))
(defun end-of-10days (date)
  (multiple-value-bind (year month day) (decode-date date)
    (cond ((<= 1 day 10) (encode-date year month 11 nil))
          ((<= 11 day 20) (encode-date year month 21 nil))
          (t (encode-date year (1+ month) 1 nil)))))
(defun finish-of-10days (date)
  (local-time- (end-of-10days date) 1))
  #|(multiple-value-bind (year month day) (decode-date date)
    (encode-date year month (cond ((<= 1 day 10) 10)
                                  ((<= 11 day 20) 20)
                                  (t (days-in-month year month))) nil)))|#

(defun start-of-month (date)
  (multiple-value-bind (year month) (decode-date date)
    (encode-date year month 1 nil)))
(defun end-of-month (date)
  (multiple-value-bind (year month) (decode-date date)
    (encode-date year (1+ month) 1 nil)))
(defun finish-of-month (date)
  (local-time- (end-of-month date) 1))
  #|(multiple-value-bind (year month) (decode-date date)
    (encode-date year month (days-in-month year month) nil)))|#

(defun start-of-quarter (date)
  (multiple-value-bind (year month)
      (%lt-day-ymd (if (integerp date) date (local-time-day date)))
    (encode-date year (1+ (* (truncate (1- month) 3) 3)) 1 nil)))
(defun end-of-quarter (date)
  (multiple-value-bind (year month)
      (%lt-day-ymd (if (integerp date) date (local-time-day date)))
    (encode-date year (1+ (* (ceiling month 3) 3)) 1 nil)))
(defun finish-of-quarter (date)
  (local-time- (end-of-quarter date) 1))
  #|(multiple-value-bind (year month)
      (%lt-day-ymd (if (integerp date) date (local-time-day date)))
    (let ((m (* (ceiling month 3) 3)))
      (encode-date year m (days-in-month year m) nil))))|#

(defun start-of-year (date)
  (encode-date (%lt-day-ymd (if (integerp date) date (local-time-day date))) 1 1 nil))
(defun end-of-year (date)
  (encode-date (%lt-day-ymd (if (integerp date) date (local-time-day date))) 13 1 nil))
(defun finish-of-year (date)
  (local-time- (end-of-year date) 1))
  ;(encode-date (%lt-day-ymd (if (integerp date) date (local-time-day date))) 12 31 nil))


(defun file-write-time (pathname)
  (local-time :universal (file-write-date pathname)
              :zone #+win32 (lt::win32-timezone) #-win32 *standard-timezone*))


#||;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; EVALUATION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(start-of-week (lt:current-date))
(finish-of-week (lt:current-date))
(start-of-10days (lt:current-date))
(finish-of-10days (lt:current-date))
(start-of-quarter @2005-01-31)
(finish-of-quarter (lt:current-date))
(finish-of-quarter @2005-01-01)
||#